/*
  COnfigurator for Gnome (COG)
  Copyright (C) 2002 Maxximum Software
   
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <libxml/tree.h>
#include <libxml/parser.h>
#include <glib.h>
#include <string.h>
#include <dirent.h>
#include "structs.h"
#include "config.h"

gboolean check_document (xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr cur)
{
  if (cur == NULL)
    {
      fprintf (stderr,"empty document\n");
      xmlFreeDoc (doc);
      return (FALSE);
    }
  /* FIXME - when namespaces are back uncomment these lines */
  /* if (ns == NULL) */
  /*   { */
  /*     fprintf (stderr, "document of the wrong type, correct namespace not found\n"); */
  /*     xmlFreeDoc (doc); */
  /*     return (FALSE); */
  /*   } */
  if (xmlStrcmp(cur->name, (const xmlChar *) "sections"))
    {
      fprintf(stderr,"document of the wrong type, root node is not sections");
      xmlFreeDoc(doc);
      return (FALSE);
    }
  
  return (TRUE);
}

cogSectionsPtr read_xml_file (GString *fname)
{
  xmlDocPtr doc;
  xmlNsPtr ns;
  xmlNodePtr cur;

  doc = xmlParseFile (fname->str);
  if (doc == NULL)
    exit (1);

  cur = xmlDocGetRootElement (doc);
  /* FIXME - when namespaces are back make this check for the correct location */
  ns = xmlSearchNsByHref(doc, cur, (const xmlChar *) "http://...");
  if (!check_document (doc, ns, cur))
    exit (1);

  /* Now, walk the tree. */
  /* First level is a list of section's */
  return parseSections (doc, ns, cur);
}

char *get_extension (char *c)
{
  GString *return_val;
  int i;

  return_val = g_string_new (NULL);

  for (i = 4 ; i > 0 ; i--) /* Get the last 4 chars of filename (ie. the extension) */
    {
      return_val = g_string_append_c (return_val, c[strlen(c)-i]);
    }

  return return_val->str;
}

int main (int argc, char *argv[])
{
  GtkWidget *main_window;
  GString *fname, *dname;
  cogSectionsPtr sections = NULL;                                          

  DIR *d = NULL;
  struct dirent *dent;

  dname = g_string_new (NULL);
  g_string_sprintf (dname, "%s/%s/%s", PACKAGE_DATA_DIR, PACKAGE, VERSION);
  d = opendir (dname->str);
  if (d == NULL)
    {
      fprintf(stderr, "problem accessing %s\n", dname->str);
      exit (1);
    }
  
  dent = readdir (d);
  while (dent!=NULL)
    {
      /* Run through all the files in dir */
      char *ext;
      if (strlen(dent->d_name) >= 5) /* filenames of interest must have at least 5 chars (ie. a.xml) */
	{
	  ext = get_extension (dent->d_name);
	  if ((strcmp (ext, ".xml") == 0 ) ||
	      (strcmp (ext, ".XML") == 0 ))
	    {
	      /* Found a file we want to read */
	      fname = g_string_new (NULL);
	      g_string_sprintf (fname, "%s/%s/%s/%s", PACKAGE_DATA_DIR, PACKAGE, VERSION, dent->d_name);
	      sections = sections_concat (sections, read_xml_file (fname));
	    }
	}
      dent = readdir (d);
    }
  closedir (d);

  if (sections == NULL)
    {
      fprintf(stderr, "No valid GConf keys found in the available XML files or no XML files found.\n");
      exit (1);
    }

  /* printSections (sections); */

  gtk_set_locale ();
  gtk_init (&argc, &argv);

  main_window = create_main_window (sections);
  gtk_widget_show (main_window);

  gtk_main ();
  
  /* Clean up everything else before quitting. */
  xmlCleanupParser();

  exit (0);
}
