/*
  COnfigurator for Gnome (COG)
  Copyright (C) 2002 Maxximum Software
   
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "structs.h"
#include "gconf-convenience.h"
#include "debug.h"

extended_keyPtr newExtended_key ()
{
  extended_keyPtr ret = NULL;
  
  /* Allocate the struct */
  ret = (extended_keyPtr) malloc(sizeof(extended_key));
  if (ret != NULL)
    memset(ret, 0, sizeof(extended_key));
    
  return (ret);
}

/* returns NULL if the GConf key is not valid */
extended_keyPtr parseExtended_key (xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr cur)
{
  gchar *gconf_key; /* needed for intermediate storage of read GConf key */
  extended_keyPtr ret = NULL;

  ret = newExtended_key ();
  if (ret == NULL)
    {
      fprintf(stderr, "out of memory\n");
      exit (1);
    }
	  
  
  /* We don't care what the top level element name is */
  cur = cur->xmlChildrenNode;
  while (cur != NULL)
    {
      if ((!xmlStrcmp(cur->name, (const xmlChar *) "gconf_key")) && (cur->ns == ns))
	{
	  /* only add to extended_key if it is a gconf_key, and it has the correct namespace... */
	  gconf_key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
	  /* ...and the gconf key exists */
	  if (gconf_key_exists (gconf_key))
	    ret->gconf_key = gconf_key;
	  else
	    debug_printf ("%s ignored because it doesn't exists in the gconf database\n", gconf_key);
	}
      if ((!xmlStrcmp(cur->name, (const xmlChar *) "short_desc")) && (cur->ns == ns))
	{
	  /* only add to extended_key if it is a short_desc, and it has the correct namespace */
	  ret->short_desc = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
	}
      if ((!xmlStrcmp(cur->name, (const xmlChar *) "long_desc")) && (cur->ns == ns))
	{
	  /* only add to extended_key if it is a short_desc, and it has the correct namespace */
	  ret->long_desc = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
	}
      cur = cur->next;                                                       
    }                   

  /* FIXME - ugly solution  */
  if (ret->gconf_key == NULL)
    return NULL;
  else
    return ret;
}

sectionPtr newSection ()
{
  sectionPtr ret = NULL;
  
  /* Allocate the struct */
  ret = (sectionPtr) malloc(sizeof(section));
  if (ret != NULL)
    memset(ret, 0, sizeof(section));                                                                                                                
    
  return (ret);
}

sectionPtr parseSection (xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr cur)
{
  xmlChar *gconf_key;

  extended_keyPtr e_key = NULL;
  sectionPtr ret = newSection ();
  if (ret == NULL)
    {
      fprintf(stderr, "out of memory\n");
      exit (1);
    }

  /* We don't care what the top level element name is */
  cur = cur->xmlChildrenNode;
  while (cur != NULL)
    {
      if ((!xmlStrcmp(cur->name, (const xmlChar *) "gconf_key")) && (cur->ns == ns))
	{
	  /* only add to section if it is a gconf_key, and it has the correct namespace... */
	  gconf_key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
	  /* ...and the gconf key exists */
	  if (gconf_key_exists (gconf_key))
	    {
	      e_key = newExtended_key ();
	      if (e_key == NULL)
		{
		  fprintf(stderr, "out of memory\n");
		  exit (1);
		}
	      e_key->gconf_key = gconf_key;
	      ret->extended_keys = g_list_append (ret->extended_keys, e_key);
	    }
	  else
	    debug_printf ("%s ignored because it doesn't exists in the gconf database\n", gconf_key);
	}
      if ((!xmlStrcmp(cur->name, (const xmlChar *) "extended_key")) && (cur->ns == ns))
	{
	  e_key = parseExtended_key (doc, ns, cur);
	  if (e_key != NULL)
	    ret->extended_keys = g_list_append (ret->extended_keys, e_key);
	}
      cur = cur->next;                                                       
    }                   

  return (ret);
}

cogSectionsPtr newCogSections ()
{
  /* Allocate the structure to be returned. */
  cogSectionsPtr ret;

  ret = (cogSectionsPtr) malloc(sizeof(cogSections));
  if (ret != NULL)
    memset(ret, 0, sizeof(cogSections));

  return (ret);
}

cogSectionsPtr parseSections (xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr cur)
{
  cogSectionsPtr sections = NULL;                                          

  sections = newCogSections ();
  if (sections == NULL)
    {
      fprintf (stderr, "out of memory\n");
      xmlFreeDoc (doc);
      exit (1);
    }

  cur = cur->xmlChildrenNode;
  while (cur != NULL)
    {
      sectionPtr cursection = NULL;

      if ((!xmlStrcmp (cur->name, (const xmlChar *) "section")) && (cur->ns == ns))
	{
	  cursection = parseSection (doc, ns, cur);
	  if (cursection != NULL)
	    {
	      cursection->title = xmlGetProp(cur, (const xmlChar *) "title");
	      sections->sections = g_list_append (sections->sections, cursection);
	    }
	}
      cur = cur->next;
    }

  return sections;
}

cogSectionsPtr sections_concat (cogSectionsPtr c1, cogSectionsPtr c2)
{
  cogSectionsPtr return_val;

  if ((c1 == NULL) && (c2 == NULL))
    return_val = NULL;
  else
    {
      return_val = newCogSections ();

      if ((c1 == NULL) && (c2 != NULL))
	return_val->sections = g_list_concat (NULL, c2->sections);
      if ((c1 != NULL) && (c2 == NULL))
	return_val->sections = g_list_concat (c1->sections, NULL);
      if ((c1 != NULL) && (c2 != NULL))
	return_val->sections = g_list_concat (c1->sections, c2->sections);
    }

  return return_val;
}

/* Quit app when window is destroyed */
static void destroy_callback (GtkWidget *window,
			      gpointer   data)
{
  gtk_main_quit ();  
}

void entry_bool_toggled_callback(GtkWidget* w, gpointer user_data)
{
  GConfClient *client;
  gboolean b;

  debug_printf ("entry_bool_toggled_callback\n");

  client = GCONF_CLIENT(g_object_get_data (G_OBJECT (w), "client"));
  b = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(w));

  gconf_client_set_bool (client, (gchar *) user_data, b, NULL);
}

void entry_float_activated_callback(GtkWidget* w, gpointer user_data)
{
  GConfClient *client;
  gchar *str;
  double d;

  debug_printf ("entry_float_activated_callback\n");

  client = GCONF_CLIENT(g_object_get_data (G_OBJECT (w), "client"));
  str = gtk_editable_get_chars(GTK_EDITABLE(w), 0, -1);
  d = atof (str);

  debug_printf ("new value to set in %s: %d\n", (gchar *) user_data, d);

  gconf_client_set_float (client, (gchar *) user_data, d, NULL);

  g_free(str);
}

void entry_int_activated_callback(GtkWidget* w, gpointer user_data)
{
  GConfClient *client;
  gchar *str;
  int i;

  debug_printf ("entry_int_activated_callback\n");

  client = GCONF_CLIENT(g_object_get_data (G_OBJECT (w), "client"));
  str = gtk_editable_get_chars(GTK_EDITABLE(w), 0, -1);
  i = atoi (str);

  debug_printf ("new value to set in %s: %i\n", (gchar *) user_data, i);

  gconf_client_set_int (client, (gchar *) user_data, i, NULL);

  g_free(str);
}

void entry_string_activated_callback(GtkWidget* w, gpointer user_data)
{
  GConfClient *client;
  gchar *str;

  debug_printf ("entry_string_activated_callback\n");

  client = GCONF_CLIENT(g_object_get_data (G_OBJECT (w), "client"));
  str = gtk_editable_get_chars(GTK_EDITABLE(w), 0, -1);

  debug_printf ("new value to set in %s: %s\n", (gchar *) user_data, str);

  gconf_client_set_string (client, (gchar *) user_data, str, NULL);
  
  g_free(str);
}

/* Remove the notification callback when the widget monitoring
 * notifications is destroyed
 */
static void configurable_widget_destroy_callback (GtkWidget *widget,
						  gpointer   data)
{
  guint notify_id;
  GConfClient *client;

  debug_printf ("configurable_widget_destroy_callback\n");

  client = g_object_get_data (G_OBJECT (widget), "client");
  notify_id = GPOINTER_TO_UINT (g_object_get_data (G_OBJECT (widget), "notify_id"));

  if (notify_id != 0)
    gconf_client_notify_remove (client, notify_id);
}

/* Notification callback for our label widgets that
 * monitor the current value of a gconf key. i.e.
 * we are conceptually "configuring" the label widgets
 */
static void configurable_widget_config_notify (GConfClient *client,
					       guint        cnxn_id,
					       GConfEntry  *entry,
					       gpointer     user_data)
{
  /* Based on the type of the entry/key we will know what type of widget */
  /* we are dealing with */
  /* This of course breaks when a key we are monitoring is deleted - FIXME */
  GtkWidget *w = user_data;
  gchar *str;

  debug_printf ("configurable_widget_config_notify\n");

  if (gconf_entry_get_value (entry) == NULL)
    {
      /* Key has been deleted */
    }
  else if (gconf_entry_get_value (entry)->type == GCONF_VALUE_FLOAT)
    {
      str = g_strdup_printf ("%f", gconf_value_get_float (gconf_entry_get_value (entry)));
      gtk_entry_set_text (GTK_ENTRY(w), str);
    }
  else if (gconf_entry_get_value (entry)->type == GCONF_VALUE_INT)
    {
      str = g_strdup_printf ("%i", gconf_value_get_int (gconf_entry_get_value (entry)));
      gtk_entry_set_text (GTK_ENTRY(w), str);
    }
  else if (gconf_entry_get_value (entry)->type == GCONF_VALUE_STRING)
    {
      gtk_entry_set_text (GTK_ENTRY (w),
			  gconf_value_get_string (gconf_entry_get_value (entry)));
    }
  else if (gconf_entry_get_value (entry)->type == GCONF_VALUE_BOOL)
    {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(w), gconf_value_get_bool (gconf_entry_get_value (entry)));
    }
  else if (gconf_entry_get_value (entry)->type == GCONF_VALUE_LIST)
    {
    }
  else if (gconf_entry_get_value (entry)->type == GCONF_VALUE_PAIR)
    {
    }
  else if (gconf_entry_get_value (entry)->type == GCONF_VALUE_SCHEMA)
    {
    }
  else if (gconf_entry_get_value (entry)->type == GCONF_VALUE_INVALID)
    {
    }
  else
    {
      /* This can't happen */
    }
}

static GtkWidget *create_configurable_widget (extended_keyPtr e_key)
{
  GtkWidget *return_widget;
  guint notify_id;
  gchar *str;
  GConfClient *client;
  GConfEntry *entry;

  gchar *config_key = e_key->gconf_key;

  debug_printf ("create_configurable_widget\n");

  /* Get the default client */
  client = gconf_client_get_default ();
  gconf_client_add_dir (client, "/", GCONF_CLIENT_PRELOAD_NONE, NULL); /* We want to monitor _all_ directories */

  entry = gconf_client_get_entry (client, config_key, NULL, TRUE, NULL);
  
  /* Note that value can be NULL (unset) or it can have
   * the wrong type! Need to check that to survive
   * gconftool --break-key
   */
  if (gconf_entry_get_value (entry) == NULL)
    {
      str = g_strdup_printf ("unset");
      return_widget = gtk_label_new (str);
    }
  else if (gconf_entry_get_value (entry)->type == GCONF_VALUE_FLOAT)
    {
      str = g_strdup_printf ("%f", gconf_value_get_float (gconf_entry_get_value (entry)));
      return_widget = gtk_entry_new ();
      gtk_entry_set_text (GTK_ENTRY(return_widget), str);
      g_signal_connect (G_OBJECT (return_widget), "activate",
			G_CALLBACK (entry_float_activated_callback),
			config_key);
    }
  else if (gconf_entry_get_value (entry)->type == GCONF_VALUE_INT)
    {
      str = g_strdup_printf ("%i", gconf_value_get_int (gconf_entry_get_value (entry)));
      return_widget = gtk_entry_new ();
      gtk_entry_set_text (GTK_ENTRY(return_widget), str);
      g_signal_connect (G_OBJECT (return_widget), "activate",
			G_CALLBACK (entry_int_activated_callback),
			config_key);
    }
  else if (gconf_entry_get_value (entry)->type == GCONF_VALUE_STRING)
    {
      return_widget = gtk_entry_new ();
      gtk_entry_set_text (GTK_ENTRY(return_widget), gconf_value_get_string (gconf_entry_get_value (entry)));
      g_signal_connect (G_OBJECT (return_widget), "activate",
			G_CALLBACK (entry_string_activated_callback),
			config_key);
    }
  else if (gconf_entry_get_value (entry)->type == GCONF_VALUE_BOOL)
    {
      return_widget = gtk_check_button_new ();
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(return_widget), gconf_value_get_bool (gconf_entry_get_value (entry)));
      g_signal_connect (G_OBJECT (return_widget), "toggled",
			G_CALLBACK (entry_bool_toggled_callback),
			config_key);
    }
  else if (gconf_entry_get_value (entry)->type == GCONF_VALUE_LIST)
    {
      str = g_strdup_printf ("list");
      return_widget = gtk_label_new (str);
    }
  else if (gconf_entry_get_value (entry)->type == GCONF_VALUE_PAIR)
    {
      str = g_strdup_printf ("pair");
      return_widget = gtk_label_new (str);
    }
  else if (gconf_entry_get_value (entry)->type == GCONF_VALUE_SCHEMA)
    {
      str = g_strdup_printf ("schema");
      return_widget = gtk_label_new (str);
    }
  else if (gconf_entry_get_value (entry)->type == GCONF_VALUE_INVALID)
    {
      str = g_strdup_printf ("invalid");
      return_widget = gtk_label_new (str);
    }
  else
    {
      str = g_strdup_printf ("this can't happen");
      return_widget = gtk_label_new (str);
    }
  
  notify_id = gconf_client_notify_add (client,
                                       config_key,
                                       configurable_widget_config_notify,
                                       return_widget,
                                       NULL, NULL);

  debug_printf ("notify_id for %s: %u\n", config_key, notify_id);
  
  /* Note that notify_id will be 0 if there was an error,
   * so we handle that in our destroy callback.
   */
  
  g_object_set_data (G_OBJECT (return_widget), "notify_id", GUINT_TO_POINTER (notify_id));
  g_object_set_data (G_OBJECT (return_widget), "client", client);

  g_signal_connect (G_OBJECT (return_widget), "destroy",
                    G_CALLBACK (configurable_widget_destroy_callback),
                    NULL);

  g_object_unref (G_OBJECT (client));

  return return_widget;
}

const gchar *extended_key_get_short_desc (extended_keyPtr e_key)
{
  if (e_key->short_desc != NULL)
    return e_key->short_desc;
  else
    return gconf_key_get_short_desc (e_key->gconf_key);
}

const gchar *extended_key_get_long_desc (extended_keyPtr e_key)
{
  if (e_key->long_desc != NULL)
    return e_key->long_desc;
  else
    return gconf_key_get_long_desc (e_key->gconf_key);
}

gboolean extended_key_is_writable (extended_keyPtr e_key)
{
  return gconf_key_is_writable (e_key->gconf_key);
}

gboolean extended_key_exists (extended_keyPtr e_key)
{
  return gconf_key_exists (e_key->gconf_key);
}

GtkWidget *create_main_window (cogSectionsPtr p)
{
  GtkWidget *main_window;
  GtkWidget *vbox;
  GtkWidget *hbuttonbox;
  GtkWidget *notebook;
  GtkWidget *close_button;
  GtkTooltips *tooltips;
  GtkWidget *scrolledwindow;
  GtkWidget *viewport;
  int i, j;

  tooltips = gtk_tooltips_new ();

  /* Add all the required widgets */
  main_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_widget_set_name (main_window, "main_window");
  gtk_window_set_title (GTK_WINDOW (main_window), "COnfigurator for Gnome (COG)");
  gtk_window_set_default_size (GTK_WINDOW (main_window), 450, 350);
  g_signal_connect (G_OBJECT (main_window), "destroy",
                    G_CALLBACK (destroy_callback), NULL);

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_widget_set_name (vbox, "vbox");
  gtk_widget_show (vbox);
  gtk_container_add (GTK_CONTAINER (main_window), vbox);

  notebook = gtk_notebook_new ();
  gtk_widget_set_name (notebook, "notebook");
  gtk_widget_show (notebook);
  gtk_box_pack_start (GTK_BOX (vbox), notebook, TRUE, TRUE, 0);
  gtk_notebook_set_scrollable (GTK_NOTEBOOK (notebook), TRUE);

  hbuttonbox = gtk_hbutton_box_new ();
  gtk_widget_set_name (hbuttonbox, "hbuttonbox");
  gtk_widget_show (hbuttonbox);
  gtk_box_pack_start (GTK_BOX (vbox), hbuttonbox, FALSE, FALSE, 0);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (hbuttonbox), GTK_BUTTONBOX_END);
  gtk_button_box_set_spacing (GTK_BUTTON_BOX (hbuttonbox), 0);
  gtk_container_set_border_width (GTK_CONTAINER (hbuttonbox), 2);

  close_button = gtk_button_new_from_stock ("gtk-close");
  gtk_widget_set_name (close_button, "close_button");
  gtk_widget_show (close_button);
  gtk_container_add (GTK_CONTAINER (hbuttonbox), close_button);
  GTK_WIDGET_SET_FLAGS (close_button, GTK_CAN_DEFAULT);
  g_signal_connect (G_OBJECT (close_button), "clicked",
		    G_CALLBACK (destroy_callback), NULL);

  /* Now run through the structures and add tabs, labels, etc. to the notebook */
  for (i = 0; i < g_list_length (p->sections); i++)
    {
      sectionPtr s = g_list_nth_data (p->sections, i);
      GtkWidget *table;
      GtkWidget *label;

      scrolledwindow = gtk_scrolled_window_new (NULL, NULL);
      gtk_widget_set_name (scrolledwindow, "scrolledwindow");
      gtk_widget_show (scrolledwindow);
      gtk_container_add (GTK_CONTAINER (notebook), scrolledwindow);
      gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
      
      viewport = gtk_viewport_new (NULL, NULL);
      gtk_widget_set_name (viewport, "viewport");
      gtk_widget_show (viewport);
      gtk_container_add (GTK_CONTAINER (scrolledwindow), viewport);
      
      table = gtk_table_new (g_list_length (s->extended_keys), 2, FALSE);
      gtk_widget_set_name (table, "table");
      gtk_widget_show (table);
      gtk_container_add (GTK_CONTAINER (viewport), table);
      gtk_container_set_border_width (GTK_CONTAINER (table), 5);
      gtk_table_set_row_spacings (GTK_TABLE (table), 5);
      gtk_table_set_col_spacings (GTK_TABLE (table), 5);
      
      /* Make notebook tab label */
      label = gtk_label_new (s->title);
      gtk_widget_set_name (label, "label");
      gtk_widget_show (label);
      gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook), gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), i), label);
      gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);

      for (j = 0; j < g_list_length (s->extended_keys); j++)
	{
	  GtkWidget *label;
	  GtkWidget *entry;

	  /* Make description label */
	  label = gtk_label_new (extended_key_get_short_desc ((extended_keyPtr) (g_list_nth_data (s->extended_keys, j))));
	  gtk_widget_set_name (label, "label");
	  gtk_widget_show (label);
	  gtk_table_attach (GTK_TABLE (table), label, 0, 1, j, j+1,
			    (GtkAttachOptions) (GTK_FILL),
			    (GtkAttachOptions) (0), 0, 0);
	  gtk_label_set_line_wrap (GTK_LABEL (label), TRUE);
	  gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
	  gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
	  gtk_widget_set_usize (label, 175, -2);

	  /* Make entry label */
	  entry = create_configurable_widget (((extended_keyPtr) (g_list_nth_data (s->extended_keys, j))));
	  gtk_widget_set_name (entry, "entry");
	  gtk_widget_show (entry);
	  gtk_table_attach (GTK_TABLE (table), entry, 1, 2, j, j+1,
			    (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
			    (GtkAttachOptions) (0), 0, 0);
	  gtk_tooltips_set_tip (tooltips, entry, extended_key_get_long_desc ((extended_keyPtr) (g_list_nth_data (s->extended_keys, j))), NULL);
	  /* If key isn't writable, then set insensitive */
	  gtk_widget_set_sensitive (entry, extended_key_is_writable ((extended_keyPtr) (g_list_nth_data (s->extended_keys, j))));
	}
    }

  return main_window;
}


void printSection (sectionPtr s)
{
  int i;
  extended_keyPtr e_key = NULL;
  GConfEntry *entry;
  GConfClient *client;

  /* Get the default client */
  client = gconf_client_get_default ();
  
  printf ("Section: %s\n", s->title);
  for (i = 0; i < g_list_length (s->extended_keys); i++)
    {
      e_key = (extended_keyPtr) (g_list_nth_data (s->extended_keys, i));
      printf ("gconf_key: %s", e_key->gconf_key);
      printf (" - type: ");
      entry = gconf_client_get_entry (client,
				      e_key->gconf_key,
				      NULL,
				      TRUE,
				      NULL);

      /* Note that value can be NULL (unset) or it can have
       * the wrong type! Need to check that to survive
       * gconftool --break-key
       */
      if (gconf_entry_get_value (entry) == NULL)
	{
	  printf ("unset\n");
	}
      else if (gconf_entry_get_value (entry)->type == GCONF_VALUE_FLOAT)
	{
	  printf ("float (%f)\n", gconf_value_get_float (gconf_entry_get_value (entry)));
	}
      else if (gconf_entry_get_value (entry)->type == GCONF_VALUE_INT)
	{
	  printf ("int (%i)\n", gconf_value_get_int (gconf_entry_get_value (entry)));
	}
      else if (gconf_entry_get_value (entry)->type == GCONF_VALUE_STRING)
	{
	  printf ("string (%s)\n", gconf_value_get_string (gconf_entry_get_value (entry)));
	}
      else if (gconf_entry_get_value (entry)->type == GCONF_VALUE_BOOL)
	{
	  printf ("bool");
	  if (gconf_value_get_bool (gconf_entry_get_value (entry)))
	    printf (" (true)\n");
	  else
	    printf (" (false)\n");
	}
      else if (gconf_entry_get_value (entry)->type == GCONF_VALUE_LIST)
	{
	  printf ("list\n");
	}
      else if (gconf_entry_get_value (entry)->type == GCONF_VALUE_PAIR)
	{
	  printf ("pair\n");
	}
      else if (gconf_entry_get_value (entry)->type == GCONF_VALUE_SCHEMA)
	{
	  printf ("schema\n");
	}
      else if (gconf_entry_get_value (entry)->type == GCONF_VALUE_INVALID)
	{
	  printf ("invalid\n");
	}
      else
	{
	  printf ("this can't happen\n");
	}


    }
  printf ("\n");

  g_object_unref (G_OBJECT (client));
}

void printSections (cogSectionsPtr p)
{
  int i;
  for (i = 0; i < g_list_length (p->sections); i++)
    {
      printSection ( (sectionPtr) (g_list_nth_data (p->sections, i)));
    }
}
